"""Constants for CGMetadata."""

# XMP packet header and footer for serializing XMP metadata.
XMP_PACKET_HEADER = '<?xpacket begin="\uFEFF" id="W5M0MpCehiHzreSzNTczkc9d"?>'
XMP_PACKET_FOOTER = '<?xpacket end="w"?>'

# Metadata types
XMP = "XMP"
EXIF = "EXIF"
IPTC = "IPTC"
TIFF = "TIFF"
GPS = "GPS"
WEBP = "WEBP"

# Metadata namespaces
MDTA = "mdta"
UDTA = "udta"
